/*********************************************************************\
*                                                                     *
*                             Fft2.java                               *
*                             ===============                         *
*                                                                     *
*                CLASE PARA CALCULO DE LA TRANSFORMADA DE             *
*                       FOURIER EN DOS DIMENSIONES                    *
*                          ===================                        *
*                                                                     *
*                Deriva de la clsica y omnipresente fft2r2.c         *
*                                                                     *
*                Sirve para un nmero de puntos que sea               *
*                potencia de 2.                                       *
*                                                                     *
*                Utiliza matrices de una sola dimensin.              *
*                                                                     *
*                                                                     *
\*********************************************************************/



package local.fourier;


import java.awt.*;

public class Fft2 {



public static void fft2r2 (int nfil, int ncol, int isign, double x[], double y[])


{



	double    a[] = new  double [nfil*ncol];
        double    b[] = new  double [nfil*ncol];
	double    seno[] = new double [nfil*ncol];
        double    cose[] = new double [nfil*ncol];
	double    rnfil, rncol, rntot, var0, fd, app;
	int      nf, nc, mf, mc, ld;
	int      inte[] = new int [1024];
	 int   i, j, ii, jj, k, ik;



	nf = nfil;   if(nfil < ncol) nf = ncol;

/*      ----   clculo de la Transformada de Fourier   ----          */

	nf = nfil / 2;
	nc = ncol / 2;
	rnfil = 1/Math.sqrt ((double)nfil);
	rncol = 1/Math.sqrt ((double)ncol);
	rntot = rnfil * rncol;
	mf = (int) (Math.log ((double)nfil) / Math.log(2) + .1); //LOGARITMOS NEPPERIOANOS
	mc = (int) (Math.log ((double)ncol) / Math.log(2) + .1);

/*      ----            Transformada por filas          ----          */

    /***************************************************\
    *   Esta parte se ha copiado de la fft.c para       *
    *   optimizar el tiempo de clculo                  *
    \***************************************************/

	fd = (double)isign * 6.283185 / (double)ncol;
	ld = ncol;
	for (i = 0; i < mc; i++)        {
		ld = ld / 2;
		inte[i] = ld;
	}
	for (i = 0; i < ncol; i++)        {
	   app = (double)i * fd;
	   cose[i] = Math.cos (app);
	   seno[i] = Math.sin (app);
	}

    /****************************************************/

	for (ii = 0; ii < nfil; ii++)   {
		k = ii * ncol;
		for (i = 0; i < nc; i++)  {
			ik = k + i;
			j  = i + nc;
			jj = k + j;
			a[i] = x[jj];
			b[i] = y[jj];
			a[j] = x[ik];
			b[j] = y[ik];
		}

		fftr2 (ncol, mc, isign, a, b, inte, seno, cose);

		for (i = 0; i < nc; i++)  {
			ik = k + i;
			j  = i + nc;
			jj = k + j;
			x[jj] = a[i] ;
			y[jj] = b[i] ;
			x[ik] = a[j] ;
			y[ik] = b[j] ;
		}
	}


/*      ----            Transformada por columnas       ----          */


    /***************************************************\
    *   Esta parte se ha copiado de la fft.c para       *
    *   optimizar el tiempo de clculo                  *
    \***************************************************/

	if ( nfil != ncol ) {
	  fd = (double)isign * 6.283185 / (double)nfil;
	  ld = nfil;
	  for (i = 0; i < mf; i++)        {
		  ld = ld / 2;
		  inte[i] = ld;
	  }
	  for (i = 0; i < nfil; i++)        {
	      app = (double)i * fd;
	      cose[i] = Math.cos (app);
	      seno[i] = Math.sin (app);
	  }
	}

    /*************************************************/
	k = nf *  ncol;
	for (jj = 0; jj < ncol; jj++)   {
		for (i = 0; i < nf; i++)  {
			j = i + nf;
			ii = i * ncol + jj;
			ik = ii + k;
			a[i] = x[ik];
			b[i] = y[ik];
			a[j] = x[ii];
			b[j] = y[ii];
		}

		fftr2 (nfil, mf, isign, a, b, inte, seno, cose);

		for (i = 0; i < nf; i++)  {
			j = i + nf;
			ii = i * ncol + jj;
			ik = ii + k;
			x[ik] = a[i] * rntot;
			y[ik] = b[i] * rntot;
			x[ii] = a[j] * rntot;
			y[ii] = b[j] * rntot;
		}
	}
	return;
}



 /********************************************************************\
 *                                                                    *
 *                        F F T   . C                                 *
 *                        ===========                                 *
 *                                                                    *
 *      Funcin para calcular la transformada de Fourier.             *
 *      Trabaja con un nmero de puntos que sea potencia de 2.        *
 *      Parmetros:                                                   *
 *              l       nmero de puntos                              *
 *              n       exponente de la potencia de 2                 *
 *              s         1 para transformada directa                 *
 *                      - 1 para transformada inversa                 *
 *              x       vector de datos ( parte real )                *
 *              y       vector de datos ( parte imaginaria )          *
 *                                                                    *
 *                                                                    *
 \********************************************************************/



public static void fftr2 (int l, int n, int s, double x[], double y[], int inte[], double seno[], double cose[])

{
	int     ld, lt, lc, lci, i, m;
	int     nt, nd, nc, ib, j, k, id=0, nci;
	double   fd, fc, qu, qd,  hu, hd;

	nd = 1;
	for (m = 1; m <= n; m++)        {
		nt = nd;
		nd = nd + nd;
		lt = l / nt;
		lc = lt / 2;
		nc = 0;
		for (ib = 0; ib < nt; ib++)    {
			lci = lt * ib;
			for (i = 0; i < lc; i++)      {
				j = i + lci;
				k = j + lc;
				qu = x[k] * cose[nc] - y[k] * seno[nc];
				qd = x[k] * seno[nc] + y[k] * cose[nc];
				x[k] = x[j] - qu;
				y[k] = y[j] - qd;
				x[j] = x[j] + qu;
				y[j] = y[j] + qd;
			}
			for (i = 1; i < n; i++) {
				id = i;
				if (inte[i] > nc) break;
				nc = nc - inte[i];
			}
                        nc = nc + inte[id];

		}
	}
	nc = 0;
	for (k = 0; k < l; k++) {
		if (nc > k)     {
			hu = x[nc];
			hd = y[nc];
			x[nc] = x[k];
			y[nc] = y[k];
			x[k] = hu;
			y[k] = hd;
		}
		for (i = 0; i < n; i++)     {
			id = i;
			if (inte[i] > nc)    break;
			nc = nc - inte[i];
		}
                nc = nc + inte[id];
	}
	return;
}





   public static void correlacioMf(int nfil, int ncol, double xr[], double xi[], double yr[], double yi[]){
      fft2r2(nfil,ncol,1,xr,xi);
      fft2r2(nfil,ncol,1,yr,yi);
      double xr0=0, yr0=0, xi0=0, yi0=0;
      for(int i=0; i<(nfil*ncol); i++){
	 xr0=xr[i]; xi0=xi[i]; yr0=yr[i]; yi0=yi[i];
	 xr[i]=xr0*yr0+xi0*yi0;
	 xi[i]=-xr0*yi0+xi0*yr0;
      }
      fft2r2(nfil,ncol,-1,xr,xi);
   }

   public static void correlacioPof(int nfil, int ncol, double xr[], double xi[], double yr[], double yi[]){
      fft2r2(nfil,ncol,1,xr,xi);
      fft2r2(nfil,ncol,1,yr,yi);
      double xr0=0, yr0=0, xi0=0, yi0=0;
      for(int i=0; i<(nfil*ncol); i++){
	 xr0=xr[i]; xi0=xi[i]; yr0=yr[i]; yi0=yi[i];
	 xr[i]=(xr0*yr0+xi0*yi0)/Math.sqrt(yr0*yr0+yi0*yi0+1e-8);
	 xi[i]=(-xr0*yi0+xi0*yr0)/Math.sqrt(yr0*yr0+yi0*yi0+1e-8);
      }
      fft2r2(nfil,ncol,-1,xr,xi);

   }

   public static void correlacioIf(int nfil, int ncol, double xr[], double xi[], double yr[], double yi[]){
      fft2r2(nfil,ncol,1,xr,xi);
      fft2r2(nfil,ncol,1,yr,yi);
      double xr0=0, yr0=0, xi0=0, yi0=0;
      for(int i=0; i<(nfil*ncol); i++){
	 xr0=xr[i]; xi0=xi[i]; yr0=yr[i]; yi0=yi[i];
	 xr[i]=(xr0*yr0+xi0*yi0)/(yr0*yr0+yi0*yi0+1e-8);
	 xi[i]=(-xr0*yi0+xi0*yr0)/(yr0*yr0+yi0*yi0+1e-8);
      }
      fft2r2(nfil,ncol,-1,xr,xi);
   }


   public static void convolucio(int nfil, int ncol, double xr[], double xi[], double yr[], double yi[]){
      fft2r2(nfil,ncol,1,xr,xi);
      fft2r2(nfil,ncol,1,yr,yi);
      double xr0=0, yr0=0, xi0=0, yi0=0;
      for(int i=0; i<(nfil*ncol); i++){
	 xr0=xr[i]; xi0=xi[i]; yr0=yr[i]; yi0=yi[i];
	 xr[i]=xr0*yr0-xi0*yi0;
	 xi[i]=xr0*yi0+xi0*yr0;
      }
      fft2r2(nfil,ncol,-1,xr,xi);
   }

/*
   public static void onaEsfericaFresnel(double x0, double y0, double z,
					 double lambda, double ampli[],
					 double partReal[], double partImag[],
					 double L, double N){


      double x=0, y=0;


      for(int i=0;i<N;i++){
	 y=(i-0.5*N)*L/N;

	 for(int j=0;j<N;j++){
	    x=(j-0.5*N)*L/N;

	    partReal[i*N+j]=(ampli[i*N+j]/z)*Math.cos(2*Math.PI*z/lambda+Math.PI*((x-x0)*(x-x0)+(y-y0)*(y-y0))/(lambda*z));
            partImag[i*N+j]=(ampli[i*N+j]/z)*Math.sin(2*Math.PI*z/lambda+Math.PI*((x-x0)*(x-x0)+(y-y0)*(y-y0))/(lambda*z));

	 }

      }

   }

*/


   public static void HologramaFresnel(double z, double lambda, double ampli[],
					 double partReal[], double partImag[],
					 double L, double N, double LR, double NR){

      /*
       * z distancia holograma pantalla
       * lambda longitud d'ona, of course!
       * ampli[] vector amb la imatge plana a holografiar
       * partReal i partImag, no comment
       * L, LR, longitud de la pantalla i de la reconstrucci en mm
       * N i NR digitalitzaci a la pantalla i ala reconstrucci
       *
       */

      double x=0, y=0, x0=0, y0=0, argument=0;

      /*
       * (x0,y0) generador de l'ona esfrica (reconstrucci)
       * (x,y) punt a la pantalla
       *
       * k, l ndexs recorrent l'holograma
       * i, j ndexs recorrent la pantalla
       *
       */

      Frame info = new Frame ("Progres clcul holograma");
      TextArea area = new TextArea ("Progres: ",380,100,0);
      info.setSize(400,120);
//      area.setBackground(Color.black);
//      area.setForeground(Color.white);
      area.setFont(new Font ("Dialog", Font.BOLD, 12));
      area.setEditable(false);
      info.add(area);
      info.show();



      for(int k=0;k<NR;k++){

	 int tpc = (int)(100*(double) k / (double)NR);
	 area.replaceRange("Progres: "+tpc+" %. Processant fila "+k+"/"+(int)NR+"    \n",0,100);

	 y0=((double) k-0.5*NR)*LR/NR;

	 for(int l=0;l<NR;l++){

	    if(ampli[k*(int)NR+l]==0) continue;
	    x0=((double) l-0.5*NR)*LR/NR;

	    for(int i=0;i<N;i++){

	       y=((double)i-0.5*N)*L/N;

	       for(int j=0;j<N;j++){

		  x=((double)j-0.5*N)*L/N;

		  argument=2*Math.PI*z/lambda+Math.PI*((x-x0)*(x-x0)+(y-y0)*(y-y0))/(lambda*z);
                  partReal[i*(int)N+j]=partReal[i*(int)N+j]+(ampli[k*(int)NR+l]/z)*Math.cos(argument);
		  partImag[i*(int)N+j]=partImag[i*(int)N+j]+(ampli[k*(int)NR+l]/z)*Math.sin(argument);

	       }

	    }
	 }
      }
   info.hide();
   }


   public static void CorbaFase
     (double partReal[], double partImag[], int mapa[], int nfil, int ncol){

	for(int i=0;i<(nfil * ncol);i++){

	 mapa[i]=(int) (254*(Math.atan2(partImag[i],partReal[i])+Math.PI)/(2*Math.PI));
	 mapa[i]=(255 << 24) | (mapa[i] << 16) | (mapa[i] << 8) |mapa[i];
      }
   }


   public static void Fresnel(int nfil, int ncol, double maxi, double mini, double partReal[], double partImag[], double intensitat[], int mapa[], int mapaPlot[], double z, double lambda, double costat){

      double pr=0, pi=0, u=0, v=0;

      fft2r2(nfil,ncol,1,partReal,partImag);

      for(int i=0; i<nfil; i++){
	 for(int j=0; j<ncol; j++){



	    u=((double)i-(double)nfil/2)/costat;
	    v=((double)j-(double)ncol/2)/costat;


	    pr=partReal[i*nfil+j]; pi=partImag[i*nfil+j];

	    partReal[i*nfil+j]=pr*Math.cos(Math.PI*lambda*z*(u*u+v*v))+pi*Math.sin(Math.PI*lambda*z*(u*u+v*v));
	    partImag[i*nfil+j]=pi*Math.cos(Math.PI*lambda*z*(u*u+v*v))-pr*Math.sin(Math.PI*lambda*z*(u*u+v*v));

	 }

      }



      fft2r2(nfil,ncol,-1,partReal,partImag);

      normaliza(nfil, ncol, maxi, mini, partReal, partImag, mapa, mapaPlot);

      for(int i=0; i< (nfil*ncol); i++)
        intensitat[i]=partReal[i]*partReal[i]+partImag[i]*partImag[i];
   }


   public static void normaliza (int nfil, int ncol, double maxi, double mini, double partReal[], double partImag[], int mapa[], int mapaPlot[]){

      double amplitud=0;

      for(int i=0; i <(nfil * ncol) ;i++){
	 amplitud=partReal[i]*partReal[i]+partImag[i]*partImag[i];
      	 if(maxi<amplitud) maxi=amplitud;
      	 if(mini>amplitud) mini=amplitud;
      }


      for(int i=0;i<(nfil * ncol);i++){
	   mapaPlot[i]=mapa[i]=(int)(254*(partReal[i]*partReal[i]+partImag[i]*partImag[i]-mini)/(maxi-mini));
	   mapa[i]=(255 << 24) | (mapa[i] << 16) | (mapa[i] << 8) |mapa[i];
      }

   }


}
