/******************************************************************************\
*                                                                              *
*                                                                              *
*              Clases y funciones para leer y grabar imgenes                  *
*                          Funciona con buffer                                 *
*                                                                              *
\******************************************************************************/



package local.imagenes;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;

import local.dialogos.OptDialogos;


//******************************************************************************


public class OptImagen extends OptVentanaNormal
{
    public BufferedImage   buffImage = null;
    public Graphics2D  g2 = null;
    public int     mapa[];
    public Image   imag = null;
    public String  nombre;
    public int     nfil, ncol, bytesPixel;
    public String fichero; //Encarni 18-3-01

//==============================================================================

//----------------------      Constructor     ----------------------------------

    public OptImagen (String titulo, int ancho, int alto)
    {
        super (titulo, ancho, alto);
        nombre = titulo;
    }

//==============================================================================

//---------------    Dibuja la imagen cada vez que es necesario  ---------------

    public void paint (Graphics g)
    {
        if (buffImage != null)
        {
            g.drawImage (buffImage, 5, 25, this);
        }
    }


// ====================================================================


  public int creaVentanaOptImagen(Frame parent, int nfil0, int ncol0){

   bytesPixel=0;
   mapa= new int[nfil0*ncol0];
   for(int i=0;i<nfil0*ncol0;i++){mapa[i]=(255 << 24) | (255 << 16) | (255 << 8) | 255;  }
   imag=createImage(new MemoryImageSource(ncol0,nfil0,mapa,0,nfil0)); //**cambio raul 21-01-02
   nombre= "";
   buffImage = new BufferedImage (ncol0, nfil0,BufferedImage.TYPE_INT_ARGB);
   g2 = buffImage.createGraphics ();
   do g2.drawImage(imag,0,0,this);
   while (buffImage.getRGB(ncol0-1,nfil0-1)==0); //** cambio
   this.setVisible(true);
   nfil=nfil0;
   ncol=ncol0;
   return 0;
  }

//==============================================================================

//--------   Lee ficheros de imgenes en formato "raw", BMP, GIF y JPG   -------

//  Si crea la imagen devuelve 1;
//  Si se cancela devuelve 0;
//  Si hay error de lectura del fichero devuelve - 1


    public int leeFicheroDeImagen (Frame parent)
    {
        Font      fuente = new Font ("Dialog", Font.BOLD, 12);

        String preguntas[] = { "N. de files",  "N. de columnes",
                     "Imatge en gris o en color (g/c)"};
        String respuestas[] = { "256", "256", "g" };


        //  Datos para el encabezamiento de BMP

        int       tipo;
        int       i, j;
        int       v1, v2, l1, l2;
        short     bitplanes = 1, bits;



        //      Crea un selector de archivos

        FileDialog selector = new FileDialog (parent, "Selector");
        // selector.setForeground (Color.blue);
        selector.setFont (fuente);
        selector.show ();

        fichero = selector.getFile ();//Encarni 18-3-01
        String dir = selector.getDirectory ();
        String dirFichero = dir + fichero;

        if ((dir == null) || (fichero == null))
             return 0;             //  Se ha cancelado

             //  Crea la imagen

        byte    datos[];
        byte    dat, datR, datG, datB;
        int     v, vR, vG, vB;
        File    file = new File (dirFichero);

        datos = new byte[3];

            //  Lee los primeros bytes del archivo

        try
        {
            FileInputStream inFile = new FileInputStream (file);
            DataInputStream dis = new DataInputStream (inFile);
            datos[0] = (byte) dis.readByte ();
            datos[1] = (byte) dis.readByte ();
            datos[2] = (byte) dis.readByte ();

            inFile.close ();
        }

        catch (EOFException evt)
        {
            OptDialogos.optEscribeFlash (parent, "Error de fi de fitxer");
            return - 1;
        }
        catch (FileNotFoundException evt)
        {
            OptDialogos.optEscribeFlash (parent, "No es troba el fitxer");
            return - 2;
        }
        catch (IOException evt)
        {
            OptDialogos.optEscribeFlash (parent, "Error de lectura");
            return - 3;
        }


            //  Para crear una imagen a partir de archivos GIF o JPG

        if (((datos[0] == -1) && (datos[1] == -40) && (datos[2] == -1))
                                                   // Es JPG
          || ((datos[0] == 71) && (datos[1] == 73) && (datos[2] == 70)))
                                                   // Es GIF
        {

                //  Genera la imagen

            imag = Toolkit.getDefaultToolkit ().getImage (dirFichero);

            do
            {
                nfil = imag.getHeight (this);
                ncol = imag.getWidth (this);
            }   while ((nfil < 0) || (ncol < 0));
            bytesPixel = 3;

                //  Crea un "buffer" para la imagen

            buffImage = new BufferedImage (ncol, nfil,
                                BufferedImage.TYPE_INT_ARGB);
            g2 = buffImage.createGraphics ();

                //  Dibuja la imagen en el "buffer"

            do
                g2.drawImage (imag, 0, 0, this);
            while
                (buffImage.getRGB (ncol-1, nfil-1) == 0);

        }

            //  Para crear una imagen a partir de archivos BMP o "raw"

        else
        {
                    //  Lee el fichero seguido y lo enva a "dato[]"

            int     numDatos;
            int     bytesLeidos;
            byte    dato[] = null;
            FileInputStream     inFile;

            try
            {
                inFile = new FileInputStream (file);
                numDatos = inFile.available ();
                dato = new byte[numDatos];
                bytesLeidos = inFile.read (dato, 0, numDatos);
                inFile.close ();
            }
            catch (EOFException evt)
            {
                OptDialogos.optEscribeFlash (parent, "Error de lectura");
                return - 1;
            }
            catch (FileNotFoundException evt)
            {
                OptDialogos.optEscribeFlash (parent, "No es troba el fitxer");
                return - 1;
            }
            catch (IOException evt)
            {
                OptDialogos.optEscribeFlash (parent, "Error de lectura");
                return - 1;
            }



            if ((datos[0] == 66) && (datos[1] == 77))   // Es BMP
            {
                    //  Pone la direccin de "headersize" en el disco que es 10
                    //  Lee el valor de "headersize" y lo pone en l2
                    //  Eso es la direccin en que empieza el "bitmap"

                l2 = (dato[10]) | (dato[11] << 8) |
                     (dato[12] << 16) | (dato[13] << 24);


                    //  Pone la direccin del n de columnas que es 18


                v1 = dato[18];
                    if (v1 < 0)     v1 += 256;
                v2 = dato[19];
                    if (v2 < 0)     v2 += 256;
                ncol = v1 | (v2 << 8);

                v1 = dato[22];
                    if (v1 < 0)     v1 += 256;
                v2 = dato[23];
                    if (v2 < 0)     v2 += 256;
                nfil = v1 | (v2 << 8);

                bitplanes = (short) dato[26];

                bits = (short)  dato[28];

                tipo = bits / 8;
                bytesPixel = tipo;


                    //  Crea un "buffer" para la imagen

                buffImage = new BufferedImage (ncol, nfil,
                                BufferedImage.TYPE_INT_ARGB);

                        //  Lee el "bitmap" para gris

                if (tipo == 1)
                {
                    for (i = nfil - 1; i >= 0; i--)
                    {
                        for (j = 0; j < ncol; j++)
                        {
                            v = (int) dato[l2];

                                    //  Lo convierte en "unsigned"

                            if (v < 0)      v += 256;

                                    //  Crea un entero de 4 bytes con un signo
                                    //  menos y las tres componentes RGB

                            v = (v << 16) | (v << 8) | v;
                            v = v | 0xff000000;

                                    //  Prepara el "buffer" de la imagen

                            buffImage.setRGB (j, i, v);

                            l2 ++;
                        }

                        while ((j % 4) != 0)
                        {
                            j ++;
                            l2++;
                        }
                    }
                }

                        //  Lee el "bitmap" para color

                if (tipo == 3)
                                    //  Imagen en color
                {
                    for (i = nfil - 1; i >= 0; i--)
                    {
                        for (j = 0; j < ncol*3; j += 3)
                        {
                                //  Lo convierte en "unsigned"
                            vR = (int) dato[l2+2];
                            if (vR < 0)     vR += 256;

                                //  Lo convierte en "unsigned"
                            vG = (int) dato[l2+1];
                            if (vG < 0)     vG += 256;

                                //  Lo convierte en "unsigned"
                            vB = (int) dato[l2];
                            if (vB < 0)     vB += 256;

                                //  Crea un entero de 4 bytes con un signo
                                //  menos y las tres componentes RGB

                            v = (vR << 16) | (vG << 8) | vB;
                            v = v | 0xff000000;

                                //  Prepara el "buffer" de la imagen

                            buffImage.setRGB (j/3, i, v);

                            l2 += 3;
                        }
                        while ((j % 4) != 0)
                        {
                            j ++;
                            l2 ++;
                        }
                    }
                }

            }
            else                            //  Es "raw"
            {

                    //  Llama a la funcin que introduce varios datos

                String dialogo = OptDialogos.optPreguntas
                                 (parent, preguntas, respuestas);

                if (dialogo == "Cancelar")
                    return 0;

                nfil = Integer.valueOf (respuestas[0]).intValue ();
                ncol = Integer.valueOf (respuestas[1]).intValue ();

                    //  Crea un "buffer" para la imagen

                buffImage = new BufferedImage (ncol, nfil,
                                BufferedImage.TYPE_INT_ARGB);

                char letra = respuestas[2].charAt (0);
                if (letra == 'g')
                    bytesPixel = 1;
                else
                    bytesPixel = 3;

                                    //  Imagen en gris

                if (bytesPixel == 1)
                {
                    for (i = 0; i < nfil; i++)
                    {
                        for (j = 0; j < ncol; j++)
                        {
                            v = (int) dato[i*ncol+j];

                                //  Lo convierte en "unsigned"

                            if (v < 0)      v += 256;

                                //  Crea un entero de 4 bytes con un signo
                                //  menos y las tres componentes RGB

                            v = (v << 16) | (v << 8) | v;
                            v = v | 0xff000000;

                                //  Prepara el "buffer" de la imagen

                            buffImage.setRGB (j, i, v);
                        }
                    }
                }
                else
                            //  Imagen en color
                {
                    for (i = 0; i < nfil; i ++)
                    {
                        for (j = 0; j < ncol; j ++)
                        {
                                //  Lo convierte en "unsigned"
                            vR = (int) dato[i*ncol*3+j*3];
                            if (vR < 0)     vR += 256;

                                //  Lo convierte en "unsigned"
                            vG = (int) dato[i*ncol*3+j*3+1];
                            if (vG < 0)     vG += 256;

                                //  Lo convierte en "unsigned"
                            vB = (int) dato[i*ncol*3+j*3+2];
                            if (vB < 0)     vB += 256;

                                //  Crea un entero de 4 bytes con un signo
                                //  menos y las tres componentes RGB

                            v = (vR << 16) | (vG << 8) | vB;
                            v = v | 0xff000000;

                                //  Prepara el "buffer" de la imagen

                            buffImage.setRGB (j, i, v);
                        }
                    }
                }
            }
        }

            //  Ajusta el tamao de la ventana y la hace visible

        this.setBounds (100, 100, ncol+10, nfil+30);  // Aade el marco
        this.setVisible (true);
        return 1;
    }



//==============================================================================

//-------------  Pregunta el formato de fichero en la grabacin   --------------


        class OptPreguntaFormatoFichero extends Dialog
        {
            Font      fuente;
            String    formato;


            public OptPreguntaFormatoFichero  (Frame parent)

            {
                super (parent, " ", true);
                setLayout (null);

                fuente = new Font ("Dialog", Font.BOLD, 12);

                    //  Construye un objeto con el texto de la pregunta

                Label pregunta = new Label ("Format del fitxer");
                pregunta.setBounds (50, 40, 280, 30);
                pregunta.setFont (fuente);
                this.add (pregunta);

                    //  Construye los botones de control

                Button  bmp = new Button ("BMP");
                bmp.addActionListener (new DialogHandler ());
                bmp.setBounds (20, 100, 50, 30);
                bmp.setFont (fuente);
                this.add (bmp);

                Button  raw = new Button ("RAW");
                raw.addActionListener (new DialogHandler ());
                raw.setBounds (90, 100, 50, 30);
                raw.setFont (fuente);
                this.add (raw);

/********
                Button  gif = new Button ("GIF");
                gif.addActionListener (new DialogHandler ());
                gif.setBounds (160, 100, 50, 30);
                gif.setFont (fuente);
                this.add (gif);

                Button  jpg = new Button ("JPG");
                jpg.addActionListener (new DialogHandler ());
                jpg.setBounds (230, 100, 50, 30);
                jpg.setFont (fuente);
                this.add (jpg);
********/
            }

            class DialogHandler extends WindowAdapter implements ActionListener
            {

                public void actionPerformed (ActionEvent e)
                {
                    formato = e.getActionCommand ();
                    hide ();
                }
            }

        }


//==============================================================================

//-------  Graba ficheros de imgenes en formato "BMP" y "RAW"    --------------


    public  int grabaFicheroDeImagen (Frame parent)
    {
        Font    fuente = new Font ("Dialog", Font.BOLD, 12);
        int     v, i, j, vR, vG, vB;

        int     numDatos = nfil * ncol * bytesPixel;
        int     bytesGrabados;
        byte    dato[] = null;


        dato = new byte[numDatos];

                //  Llama al selector de ficheros

        FileDialog selector = new FileDialog (parent, "Selector");
        //selector.setForeground (Color.blue);
        selector.setFont (fuente);
        selector.show ();
        String fichero = selector.getFile ();
        String dir = selector.getDirectory ();
        String dirFichero = dir + fichero;

        if ((dir == null) || (fichero == null))
            return 0;             //  Se ha cancelado

        File    file = new File (dirFichero);
        FileOutputStream outFile;
        DataOutputStream outData;

                //  Llama al dilogo para preguntar el formato de fichero

        OptPreguntaFormatoFichero tipoFormato =
                         new OptPreguntaFormatoFichero (parent);

        tipoFormato.setSize (300, 140);
        tipoFormato.setLocation (400, 100);
        tipoFormato.show ();


                //      Imagen en formato BMP

        if (tipoFormato.formato == "BMP")
        {
                    //  Datos para el encabezamiento de BMP

            byte      id0, id1, id2, id3;
            int       l2, filesize;
            short     res0 = 0, res1 = 0;
            int       headersize;
            int       infosize = 0x28;
            short     bitplanes = 1, bits;
            int       tipo = 3;
            int       bicompres = 0, bisizeimage = 0;
            int       bix = 0, biy = 0, biclr = 0, biclrimp = 0;

                //  Si no es imagen en color se graba la paleta,
                //  y "headersize" vale 1078 (0x436)
                //  Si es en color no se graba y "headersize" vale 54 (0x36)

            int     extremo;
            if (bytesPixel == 1)
                headersize = 1078;
            else
                headersize = 54;

                //  Clculo del tamao del fichero
                    //  Cada fila del bitmap ha de tener un nmero de bytes
                    //  mltiplo de 4. Si no lo es, se aaden ceros al final.

            int     bytesFila = ncol * bytesPixel;
            int     resto = bytesFila % 4;
            if (resto != 0)
            {
                bytesFila = (bytesFila / 4 + 1) * 4;
            }

            filesize = headersize + nfil * bytesFila;
            bits = (short) (bytesPixel * 8);
            dato = new byte[filesize + nfil * 4];

                //  Preparacin de la cabecera

            dato[0] = (byte) 'B';
            dato[1] = (byte) 'M';
            dato[2] = (byte) (filesize & 0xff);
            dato[3] = (byte) ((filesize & 0xff00) >> 8);
            dato[4] = (byte) ((filesize & 0xff0000) >> 16);
            dato[5] = (byte) ((filesize & 0xff000000) >> 24);
            dato[10] = (byte) (headersize & 0xff);
            dato[11] = (byte) ((headersize & 0xff00) >> 8);
            dato[12] = (byte) ((headersize & 0xff0000) >> 16);
            dato[13] = (byte) ((headersize & 0xff000000) >> 24);
            dato[14] = (byte) 0x28;
            dato[18] = (byte) (ncol & 0xff);
            dato[19] = (byte) ((ncol & 0xff00) >> 8);
            dato[22] = (byte) (nfil & 0xff);
            dato[23] = (byte) ((nfil & 0xff00) >> 8);
            dato[26] = (byte) 1;
            dato[28] = (byte) (bits & 0xff);


                    // Si no es en color se graba la paleta
            if (bytesPixel == 1)
            {
                    //  Cada entrada ha de tener 4 bytes, aunque uno sobra
                    //  y si es en color, las componentes van en orden inverso.
                    //  Aqu grabamos una paleta lineal.

                l2 = 54;        //  Direccin inicial de grabacin de la paleta

                for (i = 0; i < 256; i++)
                {
                    v = i;
                    if (v > 127)    v -= 256;
                    for (j = 0; j < 3; j++)
                    {
                        dato[l2 + 4 * i + j] = (byte) v;
                    }
                    dato[l2 + 4 * i + 3] = 0;
                }
            }


                    //  Preparacin del vector de datos

                        //  Las filas se graban en orden inverso
                        //  Cada una ha de tener un nmero de bytes
                        //  mltiplo de 4.
                        //  Si no lo tiene se rellena con ceros

            extremo = headersize;   //  Direccin inicial de grabacin del bitmap

            if (bytesPixel == 1)
                                            //  Imagen en gris
            {
                for (i = 0; i < nfil; i ++)
                {
                    for (j = 0; j < ncol; j ++)
                    {
                        v = buffImage.getRGB (j, nfil-i-1);
                        v = v & 0xff;
                        if (v > 127)        v -= 256;
                        dato[extremo] = (byte) v;
                        extremo ++;
                    }
                    while ((j % 4) != 0)
                    {
                        j++;
                        dato[extremo] = 0;
                        extremo ++;
                    }
                }
            }
            else
                                            //  Imagen en color
            {
                for (i = 0; i < nfil; i ++)
                {
                    for (j = 0; j < ncol*3; j += 3)
                    {
                        v = buffImage.getRGB (j/3, nfil-i-1);

                        vR = v & 0xff;
                        if (vR > 127)        vR -= 256;
                        dato[extremo] = (byte) vR;
                        extremo ++;

                        vG = v & 0xff00;
                        vG = vG >> 8;
                        if (vG > 127)        vG -= 256;
                        dato[extremo] = (byte) vG;
                        extremo ++;

                        vB = v & 0xff0000;
                        vB = vB >> 16;
                        if (vB > 127)        vB -= 256;
                        dato[extremo] = (byte) vB;
                        extremo ++;

                    }

                    while ((j % 4) != 0)
                    {
                        dato[extremo] = 0;
                        j++;
                        extremo ++;
                    }
                }
            }

                                //      Grabacin

            try
            {
                outFile = new FileOutputStream (file);
                outData = new DataOutputStream (outFile);


                        //  Graba los datos

                outData.write (dato, 0, extremo);

                        //  Ciera el fichero

                outData.close ();
                outFile.close ();
            }

                        //  Recoge las excepciones

            catch (EOFException evt)
            {
                OptDialogos.optEscribeFlash (parent, "Error de grabaci");
                return - 1;
            }
            catch (FileNotFoundException evt)
            {
                OptDialogos.optEscribeFlash (parent, "Error de grabaci");
                return - 1;
            }
            catch (IOException evt)
            {
                OptDialogos.optEscribeFlash (parent, "Error de grabaci");
                return - 1;
            }

        }

                //      Imagen en formato "raw"

        else
        {
                    //  Preparacin del vector de datos

            if (bytesPixel == 1)
                                            //  Imagen en gris
            {
                for (i = 0; i < nfil; i ++)
                {
                    for (j = 0; j < ncol; j ++)
                    {
                        v = buffImage.getRGB (j, i);
                        v = v & 0xff;
                        if (v > 127)        v -= 256;
                        dato[i*ncol+j] = (byte) v;
                    }
                }
            }
            else
                                            //  Imagen en coolor
            {
                for (i = 0; i < nfil; i ++)
                {
                    for (j = 0; j < ncol; j ++)
                    {
                        v = buffImage.getRGB (j, i);

                        vR = v & 0xff0000;
                        vR = vR >> 16;
                        if (vR > 127)        vR -= 256;
                        dato[i*ncol*3+j*3] = (byte) vR;

                        vG = v & 0xff00;
                        vG = vG >> 8;
                        if (vG > 127)        vG -= 256;
                        dato[i*ncol*3+j*3+1] = (byte) vG;

                        vB = v & 0xff;
                        if (vR > 127)        vB -= 256;
                        dato[i*ncol*3+j*3+2] = (byte) vB;
                    }
                }
            }

                                //      Grabacin

            try
            {
                outFile = new FileOutputStream (file);
                outFile.write (dato, 0, numDatos);
                outFile.close ();
            }

                        //  Recoge las excepciones

            catch (EOFException evt)
            {
                OptDialogos.optEscribeFlash (parent, "Error de grabaci");
                return - 1;
            }
            catch (FileNotFoundException evt)
            {
                OptDialogos.optEscribeFlash (parent, "Error de grabaci");
                return - 1;
            }
            catch (IOException evt)
            {
                OptDialogos.optEscribeFlash (parent, "Error de grabaci");
                return - 1;
            }
        }

        return 1;
    }

}
